#define G_LOG_DOMAIN "ap-manager"

#include "ap-ability.h"
#include "ap-skill.h"
#include "ap-weapon.h"
#include "ap-manager.h"

#include <glib.h>
#include <glib-object.h>

typedef struct _APManagerPrivate
{
    GHashTable *abilities;
    GHashTable *skills;

    GHashTable *skill_modifier_ability_bindings;
} APManagerPrivate;

struct _APManager
{
    GObject parent_instance;
};

/**
 * SECTION:ap-manager
 * @short_description: Manages communication and registration of adventure-pack
 * objects
 * @title: APManager
 * @include: adventure-pack/ap-manager.h
 * @see_also: #APAbility, #APSkill, #APAttack, #APItem, #APInventory
 *
 * Manages communication and registration of adventure-pack objects like..
 * * #APAbility
 * * #APSkill
 * * #APAttack
 * * #APItem
 *
 * # Schema # {#schema}
 * |[<!-- language="plain" -->
 * APManager
 * ├── abilities
 * │   ╰── ability
 * ├── skills
 * │   ╰── skill
 * ╰── inventory
 *     ╰── item
 * ]|
 *
 * # Purpose {#purpose}
 * ## Abilties and Skills
 *
 * #APManager automates the task of linking and updating an #APSkill to an
 * #APAbility. An #APSkill has a property named #APSkill:modifier which relates
 * to the #APAbility:id of an #APAbility. #APSkill<!-- -->s base their
 * #APSkill:score off the #APAbility:modifier of an ability, however,
 * constantly making calls and checks from an #APSkill to an #APAbility is
 * costly. Instead #APManager acts as a birdge listing for changes.
 *
 * For example if an abilties #APAbility:score changes and it modifies the
 * ability's #APAbility:modifier then a signal is emitted to #APManager.
 * #APManager then emits a `::ability-modifier-changed:(ability-id)`
 * signal to any registered #APSkill whos skill #APSKill:modifier matches the
 * ability's #APAbility:id.
 *
 * # Warnings
 * ## Singleton {#singleton-warning}
 * If you plan on use #APManager in a singleton manner by using
 * ap_manager_get_default() instead of ap_manager_new() then you should
 * NEVER unref the object with either g_object_unref() or g_clear_object().
 *
 * In addition #APManager is only singleton when used with
 * ap_manager_get_default(). Switching between using ap_manager_get_default(),
 * ap_manager_new(), or creating an %AP_TYPE_MANAGER with g_object_new() will
 * cause problems. As ap_manager_get_default() will use the same global
 * instance and the other two methods will create new #APManager objects.
 *
 * # Usage {#usage}
 * ## Complete Example
 * |[<!-- language="C" -->
 * APManager *manager = ap_manager_new ();
 * APAbility *ability;
 * APSkill *skill;
 *
 * ability = ap_ability_new ("dexterity");
 * ap_ability_add_score (ability, 10);
 * ap_manager_add_ability (manager, ability);
 *
 * skill = ap_skill_new ("stealth");
 * ap_skill_add_modifier ("dexterity");
 * ap_manager_add_skill (manager, skill);
 *
 * // The "stealth" skill's modifier will now update if the "dexterity" ability
 * // score is changed and causes the "dexterity" ability modifier to change as
 * // the "stealth" skill has added "dexterity" as it's modifier.
 *
 * ability = ap_manager_get_ability (manager, "dexterity");
 *
 * g_print ("Dexterity Ability Score: %i (%i)",
 *          ap_ability_get_score (ability)
 *          ap_ability_get_modifier (ability));
 *
 * // Returns: Dexterity Ability Score: 10 (0)
 *
 * skill = ap_manager_get_skill (manager, "stealth");
 *
 * g_print ("Stealth Modifier: %i",
 *          ap_skill_get_score (ability));
 *
 * // Returns: Stealth Modifier: 0
 *
 * ap_ability_set_score (ability, 12);
 *
 * g_print ("Stealth Modifier: %i",
 *          ap_skill_get_score (ability));
 *
 * // Returns: Stealth Modifier: 1
 *
 * // Stealth Modifier updated automatically with signals!
 *
 * g_object_unref (manager);
 * ]|
 */

G_DEFINE_TYPE_WITH_PRIVATE (APManager, ap_manager, G_TYPE_OBJECT)

/* Singleton instance */
APManager *ap_manager_instance = NULL;

enum
{
    ABILITY_REGISTERED,
    ABILITY_UNREGISTERED,
    SKILL_REGISTERED,
    SKILL_UNREGISTERED,
    WEAPON_REGISTERED,
    WEAPON_UNREGISTERED,
    NUM_SIGNALS
};

static guint signals[NUM_SIGNALS] = { 0, };

// TODO: Evaluate ENUM + new struct approach for GHashTable containing all binds
typedef enum _ApManagerBindType
{
    ABILITY_SCORE,
    ABILITY_MODIFIER,
    SKILL_SCORE,
    SKILL_MODIFIER
} ApManagerBindType;

typedef struct _APManagerBind
{
    gchar *name;
    //APManagerBindType type;

    gpointer source;
    gpointer target;
    GBinding *bind;
} APManagerBind;

// TODO: Find a better solution?
static void
on_manager_skill_modifier_changed (GObject    *source,
                                   GParamSpec *pspec,
                                   gpointer    user_data)
{
    APManager *self = AP_MANAGER (user_data);
    APManagerPrivate *priv = ap_manager_get_instance_private (self);
    APSkill *skill = AP_SKILL (source);
    gchar *name = ap_skill_get_name (skill);
    gchar *modifier = ap_skill_get_modifier (skill);
    APAbility *ability;

    if (modifier == NULL)
        g_hash_table_remove (priv->skill_modifier_ability_bindings, name);
    else
        ability = ap_manager_get_ability (self, modifier);

    if (ability)
    {
        GBinding *new_bind;

        new_bind =
            g_object_bind_property (ability, "modifier",
                                    skill, "score",
                                    G_BINDING_DEFAULT | G_BINDING_SYNC_CREATE);

        g_hash_table_replace (priv->skill_modifier_ability_bindings,
                              name,
                              new_bind);
    }

    g_free (modifier);
}

static gboolean
register_ability (APManager *self,
                  APAbility *ability)
{
    g_assert (AP_IS_MANAGER (self));
    g_assert (AP_IS_ABILITY (ability));

    APManagerPrivate *priv = ap_manager_get_instance_private (self);
    gchar *name = ap_ability_get_name (ability);

    g_debug ("Attempting to register ability with name \"%s\".", name);

    if (name == NULL)
    {
        g_warning ("Ability not registered, ability must have a name to be registered.");
        return FALSE;
    }

    if (g_hash_table_contains (priv->abilities, name))
    {
        g_warning ("Ability not registered, ability with name \"%s\" already exists.", name);
        return FALSE;
    }

    g_hash_table_insert (priv->abilities, name, ability);

    if (!g_hash_table_contains (priv->abilities, name))
    {
        g_warning ("Unable to add ability with name \"%s\" to tracker.", name);
        return FALSE;
    }

    g_object_ref (ability);

    g_signal_emit (self, signals[ABILITY_REGISTERED], 0, ability);

    g_debug ("Ability \"%s\" registered.", name);

    return TRUE;
}

static gboolean
unregister_ability (APManager *self,
                    APAbility *ability)
{
    g_assert (AP_IS_MANAGER (self));
    g_assert (AP_IS_ABILITY (ability));

    APManagerPrivate *priv = ap_manager_get_instance_private (self);
    gchar *name = ap_ability_get_name (ability);

    g_debug ("Attempting to unregister ability with name \"%s\".", name);

    //TODO: Remove signals

    g_hash_table_remove (priv->abilities, name);

    g_signal_emit (self, signals[ABILITY_UNREGISTERED], 0, ability);

    g_debug ("Ability \"%s\" unregistered.", name);

    g_free (name);

    return TRUE;
}

static gboolean
register_skill (APManager *self,
                APSkill   *skill)
{
    g_assert (AP_IS_MANAGER (self));
    g_assert (AP_IS_SKILL (skill));

    APManagerPrivate *priv = ap_manager_get_instance_private (self);
    gchar *name = ap_skill_get_name (skill);
    gchar *modifier = ap_skill_get_modifier (skill);

    g_debug ("Attempting to register skill with name \"%s\".", name);

    if (name == NULL)
    {
        g_warning ("Skill not registered, skill must have a name to be registered.");
        return FALSE;
    }

    if (g_hash_table_contains (priv->skills, name))
    {
        g_warning ("Skill not registered, skill with name \"%s\" already exists.", name);
        return FALSE;
    }

    g_hash_table_insert (priv->skills, name, skill);

    if (!g_hash_table_contains (priv->skills, name))
    {
        g_debug ("Unable to add skill with name \"%s\" to tracker.", name);
        return FALSE;
    }

    g_object_ref (skill);

    if (modifier == NULL)
    {
        g_warning ("Skill \"%s\" does not have a modifier, while not required it is recommended to set a modifier before adding it to the manager.", name);
    }

    g_signal_connect (skill,
                      "notify::modifier",
                      G_CALLBACK (on_manager_skill_modifier_changed),
                      self);

    if (modifier)
        g_object_notify (G_OBJECT (skill), "modifier");

    g_signal_emit (self, signals[SKILL_REGISTERED], 0, skill);

    g_debug ("Skill \"%s\" registered.", name);

    g_free (modifier);

    return TRUE;
}

static gboolean
unregister_skill (APManager *self,
                  APSkill   *skill)
{
    g_assert (AP_IS_MANAGER (self));
    g_assert (AP_IS_SKILL (skill));

    APManagerPrivate *priv = ap_manager_get_instance_private (self);
    gchar *name = ap_skill_get_name (skill);

    g_debug ("Attempting to unregister skill with name \"%s\".", name);

    //TODO: Remove signals

    g_hash_table_remove (priv->skills, name);

    g_signal_emit (self, signals[SKILL_UNREGISTERED], 0, skill);

    g_debug ("Skill \"%s\" unregistered.", name);

    g_free (name);

    return TRUE;
}

static void
ap_manager_dispose (GObject *object)
{
    APManager *self = AP_MANAGER (object);

    ap_manager_remove_all (self);

    G_OBJECT_CLASS (ap_manager_parent_class)->dispose (object);
}

static void
ap_manager_finalize (GObject *object)
{
    APManager *self = AP_MANAGER (object);
    APManagerPrivate *priv = ap_manager_get_instance_private (self);

    if (g_hash_table_size (priv->abilities))
        g_warning ("Not all abilities have been removed.");

    if (g_hash_table_size (priv->skills))
        g_warning ("Not all skills have been removed.");

    g_clear_pointer (&priv->abilities, g_hash_table_unref);
    g_clear_pointer (&priv->skills, g_hash_table_unref);

    G_OBJECT_CLASS (ap_manager_parent_class)->finalize (object);
}

static void
ap_manager_init (APManager *self)
{
    APManagerPrivate *priv = ap_manager_get_instance_private (self);

    priv->abilities = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
    priv->skills = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);

    priv->skill_modifier_ability_bindings = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify) g_binding_unbind);
}

static void
ap_manager_class_init (APManagerClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->dispose = ap_manager_dispose;
    object_class->finalize = ap_manager_finalize;

    /**
     * APManager::ability-registered:
     * @manager: an #APManager
     * @ability: an #APAbility
     *
     * The ::ability-registered signal is emitted each time an #APAbility is
     * registered.
     */
    signals[ABILITY_REGISTERED] =
        g_signal_new ("ability-registered",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      AP_TYPE_ABILITY);

    /**
     * APManager::ability-unregistered:
     * @manager: an #APManager
     * @ability: an #APAbility
     *
     * The ::ability-unregistered signal is emitted each time an #APAbility is
     * unregistered.
     */
    signals[ABILITY_UNREGISTERED] =
        g_signal_new ("ability-unregistered",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      AP_TYPE_ABILITY);

    /**
     * APManager::skill-registered:
     * @manager: an #APManager
     * @skill: an #APSkill
     *
     * The ::ability-registered signal is emitted each time an #APSkill is
     * registered.
     */
    signals[SKILL_REGISTERED] =
        g_signal_new ("skill-registered",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      AP_TYPE_SKILL);

    /**
     * APManager::skill-unregistered:
     * @manager: an #APManager
     * @skill: an #APSkill
     *
     * The ::ability-unregistered signal is emitted each time an #APSkill is
     * unregistered.
     */
    signals[SKILL_UNREGISTERED] =
        g_signal_new ("skill-unregistered",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_FIRST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      1,
                      AP_TYPE_SKILL);
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
/**
 * ap_manager_new:
 *
 * Creates a new #APManager which manages the connections between
 * #APAbility<!-- -->s, #APSkill<!-- -->s, #APAttack<!-- -->s,
 * #APItem<!-- -->s, etc... #APManager registers and unregisters the various
 * signals for communication between them.
 *
 * # Usage
 * |[<!-- language="C" -->
 * APManager *manager;
 *
 * manager = ap_manager_new ();
 *
 * // Do some stuff...
 *
 * g_object_unref (manager);
 * ]|
 *
 * Returns: (transfer full): an #APManager instance
 */
APManager *
ap_manager_new (void)
{
    return g_object_new (AP_TYPE_MANAGER,
                         NULL);
}

/**
 * ap_manager_get_default:
 *
 * Returns the default singleton #APManager instance, if a global #APManager
 * singleton instance does not exist then one is created.
 *
 * # Warning
 * If you use ap_manager_get_default() at all then you are agreeing to use
 * #APManager in singleton mode. See [here][singleton-warning] for more
 * information.
 *
 * # Usage
 * |[<!-- language="C" -->
 * APManager *manager;
 *
 * manager = ap_manager_get_default ();
 * ]|
 *
 * Returns: (transfer none): the singleton #APManager instance
 */
APManager *
ap_manager_get_default (void)
{
    if (!ap_manager_instance)
        ap_manager_instance = ap_manager_new ();

    return ap_manager_instance;
}

/**
 * ap_manager_add_ability:
 * @self: an #APManager
 * @ability: an #APAbility
 *
 * Adds an #APAbility to the #APManager.
 *
 * Returns: (gboolean): %TRUE if @ability has been added
 */
gboolean
ap_manager_add_ability (APManager *self,
                        APAbility *ability)
{
    g_assert (AP_IS_MANAGER (self));
    g_assert (AP_IS_ABILITY (ability));

    return register_ability (self, ability);
}

/**
 * ap_manager_get_ability:
 * @self: an #APManager
 * @name: the name of an #APAbility
 *
 * Finds and returns an ability with @name, can return %NULL if no ability with
 * @name is found.
 *
 * Returns: (transfer none) (nullable): an #APAbility
 */
APAbility *
ap_manager_get_ability (APManager *self,
                        gchar     *name)
{
    APManagerPrivate *priv;

    g_assert (AP_IS_MANAGER (self));

    priv = ap_manager_get_instance_private (self);

    return g_hash_table_lookup (priv->abilities, name);
}

/**
 * ap_manager_remove_ability:
 * @self: an #APManager
 * @name: the name of an #APAbility
 *
 * Removes an #APAbility from the #APManager.
 *
 * Returns: (gboolean): %TRUE if the #APAbility with @name was found and removed
 *          or if the #APAbility is not currently tracked by #APManager
 */
gboolean
ap_manager_remove_ability (APManager *self,
                           gchar     *name)
{
    APAbility *ability;

    g_assert (AP_IS_MANAGER (self));

    ability = ap_manager_get_ability (self, name);

    if (ability == NULL)
        return TRUE;

    return unregister_ability (self, ability);
}

/**
 * ap_manager_add_skill:
 * @self: an #APManager
 * @skill: an #APSkill
 *
 * Adds an #APSkill to the #APManager.
 *
 * Returns: (gboolean): %TRUE if @skill has been added
 */
gboolean
ap_manager_add_skill (APManager *self,
                      APSkill   *skill)
{
    g_assert (AP_IS_MANAGER (self));
    g_assert (AP_IS_SKILL (skill));

    return register_skill (self, skill);
}

/**
 * ap_manager_get_skill:
 * @self: an #APManager
 * @name: the name of an #APSkill
 *
 * Finds and returns a skill with @name, can return %NULL if no skill with
 * @name is found.
 *
 * Returns: (transfer none) (nullable): an #APSkill
 */
APSkill *
ap_manager_get_skill (APManager *self,
                      gchar     *name)
{
    APManagerPrivate *priv;

    g_assert (AP_IS_MANAGER (self));

    priv = ap_manager_get_instance_private (self);

    return g_hash_table_lookup (priv->skills, name);
}

/**
 * ap_manager_remove_skill:
 * @self: an #APManager
 * @name: the name of an #APSkill
 *
 * Removes an #APSkill from the #APManager.
 *
 * Returns: (gboolean): %TRUE if the #APSkill with @name was found and removed
 *          or if the #APSkill is not currently tracked by #APManager
 */
gboolean
ap_manager_remove_skill (APManager *self,
                         gchar     *name)
{
    APSkill *skill;

    g_assert (AP_IS_MANAGER (self));

    skill = ap_manager_get_skill (self, name);

    if (skill == NULL)
        return TRUE;

    return unregister_skill (self, skill);
}

/**
 * ap_manager_get_abilities:
 * @self: an #APManager
 *
 * Returns a #GList containing every #APAbility registered.
 *
 * Returns: (transfer container): a #GList containing all the abilities. The
 *          content of the list is owned by #APManager and should not be
 *          modified or freed. Use g_list_free() when done using the list.
 */
GList *
ap_manager_get_abilities (APManager *self)
{
    APManagerPrivate *priv;

    g_assert (AP_IS_MANAGER (self));

    priv = ap_manager_get_instance_private (self);

    return g_hash_table_get_values (priv->abilities);
}

/**
 * ap_manager_remove_abilities:
 * @self: an #APManager
 *
 * Removes all abilities from @self.
 */
void
ap_manager_remove_abilities (APManager *self)
{
    GList *abilities;
    GList *list;

    g_assert (AP_IS_MANAGER (self));

    abilities = ap_manager_get_abilities (self);

    for (list = abilities; list != NULL; list = list->next)
    {
        g_assert (AP_IS_ABILITY (list->data));

        unregister_ability (self, list->data);
    }

    g_list_free (abilities);
}

/**
 * ap_manager_get_skills:
 * @self: an #APManager
 *
 * Returns a #GList containing every #APSkill registered.
 *
 * Returns: (transfer container): a #GList containing all the skills. The
 *          content of the list is owned by #APManager and should not be
 *          modified or freed. Use g_list_free() when done using the list.
 */
GList *
ap_manager_get_skills (APManager *self)
{
    APManagerPrivate *priv;

    g_assert (AP_IS_MANAGER (self));

    priv = ap_manager_get_instance_private (self);

    return g_hash_table_get_values (priv->skills);
}

/**
 * ap_manager_remove_skills:
 * @self: an #APManager
 *
 * Removes all skills from @self.
 */
void
ap_manager_remove_skills (APManager *self)
{
    GList *skills;
    GList *list;

    g_assert (AP_IS_MANAGER (self));

    skills = ap_manager_get_skills (self);

    for (list = skills; list != NULL; list = list->next)
    {
        g_assert (AP_IS_SKILL (list->data));

        unregister_skill (self, list->data);
    }

    g_list_free (skills);
}

/**
 * ap_manager_remove_all:
 * @self: an #APManager
 *
 * Removes all #APAbilitys, #APSkills, etc.. from the #APManager.
 */
void
ap_manager_remove_all (APManager *self)
{
    g_assert (AP_IS_MANAGER (self));

    ap_manager_remove_abilities (self);
    ap_manager_remove_skills (self);
}
