#include "test-manager.h"

#include <adventure-pack/ap-ability.h>
#include <adventure-pack/ap-skill.h>
#include <adventure-pack/ap-manager.h>

#include <glib.h>

void
manager_fixture_set_up (ManagerFixture *fixture,
                        gconstpointer   user_data)
{
    fixture->manager = ap_manager_new ();

}

void
manager_fixture_tear_down (ManagerFixture *fixture,
                           gconstpointer   user_data)
{
    g_clear_object (&fixture->manager);
}

void
test_manager_ability_add (ManagerFixture *fixture,
                          gconstpointer   user_data)
{
    APAbility *ability;
    APAbility *ability_from_manager;
    gboolean retval;

    ability = ap_ability_new_with_name ("Dexterity");

    retval = ap_manager_add_ability (fixture->manager, ability);

    g_assert (retval == TRUE);

    ability_from_manager = ap_manager_get_ability (fixture->manager, "Dexterity");

    g_assert (ability_from_manager == ability);
}

void
test_manager_ability_add_with_no_name (ManagerFixture *fixture,
                                       gconstpointer   user_data)
{
    APAbility *ability;

    ability = ap_ability_new ();

    g_test_expect_message ("ap-manager",
                           G_LOG_LEVEL_WARNING,
                           "Ability not registered, ability must have a name to be registered.");

    ap_manager_add_ability (fixture->manager, ability);

    g_test_assert_expected_messages ();
}

void
test_manager_ability_add_with_duplicate_name (ManagerFixture *fixture,
                                              gconstpointer   user_data)
{
    APAbility *ability;
    gboolean retval;

    ability = ap_ability_new_with_name ("Dexterity");
    retval = ap_manager_add_ability (fixture->manager, ability);
    g_assert (retval == TRUE);

    ability = ap_ability_new_with_name ("Dexterity");

    g_test_expect_message ("ap-manager",
                           G_LOG_LEVEL_WARNING,
                           "Ability not registered, ability with name \"*\" already exists.");

    ap_manager_add_ability (fixture->manager, ability);

    g_test_assert_expected_messages ();
}

void
test_manager_ability_get (ManagerFixture *fixture,
                          gconstpointer   user_data)
{
    APAbility *ability;
    APAbility *ability_from_manager;

    ability = ap_ability_new ();

    ap_ability_set_name (ability, "Dexterity");

    ap_manager_add_ability (fixture->manager, ability);

    ability_from_manager = ap_manager_get_ability (fixture->manager, "Dexterity");

    g_assert (AP_IS_ABILITY (ability_from_manager));
    g_assert (ability == ability_from_manager);
}

void
test_manager_skill_add (ManagerFixture *fixture,
                        gconstpointer   user_data)
{
    APSkill *skill;
    gboolean retval;

    skill = ap_skill_new_with_name ("Acrobatics");

    g_test_expect_message ("ap-manager",
                           G_LOG_LEVEL_WARNING,
                           "Skill \"*\" does not have a modifier, while not required it is recommended to set a modifier before adding it to the manager.");

    retval = ap_manager_add_skill (fixture->manager, skill);

    g_test_assert_expected_messages ();

    g_assert (retval == TRUE);
}

void
test_manager_skill_add_with_no_name (ManagerFixture *fixture,
                                     gconstpointer   user_data)
{
    APSkill *skill;

    skill = ap_skill_new ();

    g_test_expect_message ("ap-manager",
                           G_LOG_LEVEL_WARNING,
                           "Skill not registered, skill must have a name to be registered.");

    ap_manager_add_skill (fixture->manager, skill);

    g_test_assert_expected_messages ();
}

void
test_manager_skill_add_with_duplicate_name (ManagerFixture *fixture,
                                            gconstpointer   user_data)
{
    APSkill *skill;

    skill = ap_skill_new_with_name ("Acrobatics");
    ap_skill_set_modifier (skill, "Dexterity");
    ap_manager_add_skill (fixture->manager, skill);

    skill = ap_skill_new_with_name ("Acrobatics");
    ap_skill_set_modifier (skill, "Dexterity");

    g_test_expect_message ("ap-manager",
                           G_LOG_LEVEL_WARNING,
                           "Skill not registered, skill with name \"*\" already exists.");

    ap_manager_add_skill (fixture->manager, skill);

    g_test_assert_expected_messages ();
}

void
test_manager_skill_get (ManagerFixture *fixture,
                        gconstpointer   user_data)
{
    APSkill *skill;
    APSkill *skill_from_manager;

    skill = ap_skill_new_with_name ("Acrobatics");

    ap_skill_set_modifier (skill, "Dexterity");

    ap_manager_add_skill (fixture->manager, skill);

    skill_from_manager = ap_manager_get_skill (fixture->manager, "Acrobatics");

    g_assert (AP_IS_SKILL (skill_from_manager));
    g_assert (skill == skill_from_manager);
}

void
test_manager_remove_ability (ManagerFixture *fixture,
                             gconstpointer   user_data)
{
    APAbility *ability;

    ability = g_object_new (AP_TYPE_ABILITY,
                            "name", "Dexterity",
                            "score", 10,
                            NULL);
    ap_manager_add_ability (fixture->manager, ability);

    g_assert (ap_manager_get_ability (fixture->manager, "Dexterity") == ability);

    ap_manager_remove_ability (fixture->manager, "Dexterity");

    g_assert (ap_manager_get_ability (fixture->manager, "Dexterity") == NULL);
}

void
test_manager_remove_abilities (ManagerFixture *fixture,
                               gconstpointer   user_data)
{
    APAbility *ability1;
    APAbility *ability2;
    APAbility *ability3;

    ability1 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Dexterity",
                             "score", 10,
                             NULL);
    ap_manager_add_ability (fixture->manager, ability1);

    ability2 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Strength",
                             "score", 10,
                             NULL);
    ap_manager_add_ability (fixture->manager, ability2);

    ability3 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Wisdom",
                             "score", 10,
                             NULL);
    ap_manager_add_ability (fixture->manager, ability3);

    g_assert (ap_manager_get_ability (fixture->manager, "Dexterity") == ability1);
    g_assert (ap_manager_get_ability (fixture->manager, "Strength") == ability2);
    g_assert (ap_manager_get_ability (fixture->manager, "Wisdom") == ability3);

    ap_manager_remove_abilities (fixture->manager);

    g_assert (ap_manager_get_ability (fixture->manager, "Dexterity") == NULL);
    g_assert (ap_manager_get_ability (fixture->manager, "Strength") == NULL);
    g_assert (ap_manager_get_ability (fixture->manager, "Wisdom") == NULL);
}

void
test_manager_remove_skill (ManagerFixture *fixture,
                           gconstpointer   user_data)
{
    APSkill *skill;

    skill = g_object_new (AP_TYPE_SKILL,
                          "name", "Acrobatics",
                          "modifier", "Dexterity",
                          NULL);
    ap_manager_add_skill (fixture->manager, skill);

    g_assert (ap_manager_get_skill (fixture->manager, "Acrobatics") == skill);

    ap_manager_remove_skill (fixture->manager, "Acrobatics");

    g_assert (ap_manager_get_skill (fixture->manager, "Acrobatics") == NULL);

}

void
test_manager_remove_skills (ManagerFixture *fixture,
                            gconstpointer   user_data)
{
    APSkill *skill1;
    APSkill *skill2;
    APSkill *skill3;

    skill1 = g_object_new (AP_TYPE_SKILL,
                             "name", "Acrobatics",
                             "modifier", "Dexterity",
                             NULL);
    ap_manager_add_skill (fixture->manager, skill1);

    skill2 = g_object_new (AP_TYPE_SKILL,
                             "name", "Animal Handling",
                             "modifier", "Wisdom",
                             NULL);
    ap_manager_add_skill (fixture->manager, skill2);

    skill3 = g_object_new (AP_TYPE_SKILL,
                             "name", "Athletics",
                             "modifier", "Strength",
                             NULL);
    ap_manager_add_skill (fixture->manager, skill3);

    g_assert (ap_manager_get_skill (fixture->manager, "Acrobatics") == skill1);
    g_assert (ap_manager_get_skill (fixture->manager, "Animal Handling") == skill2);
    g_assert (ap_manager_get_skill (fixture->manager, "Athletics") == skill3);

    ap_manager_remove_skills (fixture->manager);

    g_assert (ap_manager_get_skill (fixture->manager, "Acrobatics") == NULL);
    g_assert (ap_manager_get_skill (fixture->manager, "Animal Handling") == NULL);
    g_assert (ap_manager_get_skill (fixture->manager, "Athletics") == NULL);
}

void
test_manager_remove_all (ManagerFixture *fixture,
                         gconstpointer   user_data)
{

    APAbility *ability1;
    APAbility *ability2;
    APAbility *ability3;
    APSkill *skill1;
    APSkill *skill2;
    APSkill *skill3;

    ability1 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Dexterity",
                             "score", 10,
                             NULL);
    ap_manager_add_ability (fixture->manager, ability1);

    ability2 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Strength",
                             "score", 10,
                             NULL);
    ap_manager_add_ability (fixture->manager, ability2);

    ability3 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Wisdom",
                             "score", 10,
                             NULL);
    ap_manager_add_ability (fixture->manager, ability3);

    skill1 = g_object_new (AP_TYPE_SKILL,
                             "name", "Acrobatics",
                             "modifier", "Dexterity",
                             NULL);
    ap_manager_add_skill (fixture->manager, skill1);

    skill2 = g_object_new (AP_TYPE_SKILL,
                             "name", "Animal Handling",
                             "modifier", "Wisdom",
                             NULL);
    ap_manager_add_skill (fixture->manager, skill2);

    skill3 = g_object_new (AP_TYPE_SKILL,
                             "name", "Athletics",
                             "modifier", "Strength",
                             NULL);
    ap_manager_add_skill (fixture->manager, skill3);

    g_assert (ap_manager_get_ability (fixture->manager, "Dexterity") == ability1);
    g_assert (ap_manager_get_ability (fixture->manager, "Strength") == ability2);
    g_assert (ap_manager_get_ability (fixture->manager, "Wisdom") == ability3);

    g_assert (ap_manager_get_skill (fixture->manager, "Acrobatics") == skill1);
    g_assert (ap_manager_get_skill (fixture->manager, "Animal Handling") == skill2);
    g_assert (ap_manager_get_skill (fixture->manager, "Athletics") == skill3);

    ap_manager_remove_all (fixture->manager);

    g_assert (ap_manager_get_ability (fixture->manager, "Dexterity") == NULL);
    g_assert (ap_manager_get_ability (fixture->manager, "Strength") == NULL);
    g_assert (ap_manager_get_ability (fixture->manager, "Wisdom") == NULL);

    g_assert (ap_manager_get_skill (fixture->manager, "Acrobatics") == NULL);
    g_assert (ap_manager_get_skill (fixture->manager, "Animal Handling") == NULL);
    g_assert (ap_manager_get_skill (fixture->manager, "Athletics") == NULL);
}

void
test_manager_signal_ability_score_changed_skill_score_changed (ManagerFixture *fixture,
                                                               gconstpointer   user_data)
{
    APAbility *ability;
    APSkill *skill;

    ability = g_object_new (AP_TYPE_ABILITY,
                            "name", "Dexterity",
                            "score", 10,
                            NULL);
    ap_manager_add_ability (fixture->manager, ability);

    skill = g_object_new (AP_TYPE_SKILL,
                          "name", "Acrobatics",
                          "modifier", "Dexterity",
                          NULL);
    ap_manager_add_skill (fixture->manager, skill);

    g_assert (ap_skill_get_score (skill) == 0);

    ap_ability_set_score (ability, 12);

    g_assert (ap_skill_get_score (skill) == 1);
}

void
test_manager_signal_skill_modifier_changed (ManagerFixture *fixture,
                                            gconstpointer   user_data)
{
    APAbility *ability;
    APSkill *skill;
    gint modifier1, modifier2, modifier3;

    // NOTE:
    // We don't care if the modifier is generated correctly, ex. ability scores
    // of 10~11 = modifier of 0, ability scores of 12~13 = modifier of 1, etc..
    //
    // We have test for that, for that reason we use ap_ability_get_modifier()
    // just to make sure #APManager is sending the right signals, a.k.a the
    // right values.

    // Add Dexterity ability
    ability = ap_ability_new_with_name ("Dexterity");
    ap_ability_set_score (ability, 10);
    modifier1 = ap_ability_get_modifier (ability);
    ap_manager_add_ability (fixture->manager, ability);

    // Add Strength ability
    ability = ap_ability_new_with_name ("Strength");
    ap_ability_set_score (ability, 12);
    modifier2 = ap_ability_get_modifier (ability);
    ap_manager_add_ability (fixture->manager, ability);

    // Add Acrobatics skill that uses "Strength" ability
    skill = ap_skill_new_with_name ("Acrobatics");
    ap_skill_set_modifier (skill, "Dexterity");
    ap_manager_add_skill (fixture->manager, skill);

    // Current "Acrobatics" skill should have a modifier of 0 because an ability
    // score of 10 returns a modifier of 0
    g_assert (ap_skill_get_score (skill) == modifier1);

    // Switch skill to be based of "Strength" ability
    ap_skill_set_modifier (skill, "Strength");

    // Current "Acrobatics" skill should have a modifier of 1 because an ability
    // score of 12 returns a modifier of 1
    g_assert (ap_skill_get_score (skill) == modifier2);

    // Get "Strength" ability as we are currently using it and up it's score to
    // 14. This simulates a user increasing their ability score and changes
    // need to propagate to skills that use "Strength" as a modifier.
    ability = ap_manager_get_ability (fixture->manager, "Strength");
    ap_ability_set_score (ability, 14);
    modifier3 = ap_ability_get_modifier (ability);

    // Current "Acrobatics" skill should have a modifier of 2 because an ability
    // score of 14 returns a modifier of 2
    g_assert (ap_skill_get_score (skill) == modifier3);
}

void
test_manager_check_skill_for_lingering_bind_with_ability (ManagerFixture *fixture,
                                                          gconstpointer   user_data)
{
    APAbility *ability1;
    APAbility *ability2;
    APAbility *manager_ability;
    APSkill *skill;

    ability1 = g_object_new (AP_TYPE_ABILITY,
                            "name", "Dexterity",
                            "score", 10,
                            NULL);
    ap_manager_add_ability (fixture->manager, ability1);

    ability2 = g_object_new (AP_TYPE_ABILITY,
                            "name", "Strength",
                            "score", 12,
                            NULL);
    ap_manager_add_ability (fixture->manager, ability2);

    skill = g_object_new (AP_TYPE_SKILL,
                          "name", "Acrobatics",
                          "modifier", "Dexterity",
                          NULL);
    ap_manager_add_skill (fixture->manager, skill);

    // Connecting to an existing ability should get that's ability modifier
    // and update the Skills score
    g_assert (ap_skill_get_score (skill) == 0);

    ap_skill_set_modifier (skill, "Strength");

    // Switching skill should issue new notify:modifier & get updated
    // #APAbility:score
    g_assert (ap_skill_get_score (skill) == 1);

    manager_ability = ap_manager_get_ability (fixture->manager, ap_skill_get_modifier (skill));

    // Skill should be returning new updated "modifier"
    g_assert (manager_ability == ability2);

    ap_ability_set_score (ability1, 14);

    // Updating the old ability connected to the Skill should **NOT** update
    // the score!
    g_assert (ap_skill_get_score (skill) == 1);
}
